@@ -19,6 +19,8 @@ from .logging import *
 from .search import *
 from .evolution import *
 from .surrogate import *
+from .cmame import *
+from .sampling import *
 
 
 # MODELINE	"{{{1
