@@ -17,7 +17,7 @@
 
 All algorithms must be registered to the default qdpy.base.Registry `qdpy.base.registry` if they are to be easily accessed or created by factories. It can be done by decorating QD algorithms classes with `@registry.register`. """
 
-__all__ = ["partial", "QDAlgorithmLike", "QDAlgorithm", "Lambda", "AlgWithBatchSuggestions", "Sq"]
+__all__ = ["partial", "QDAlgorithmLike", "QDAlgorithm", "Lambda", "AlgWithBatchSuggestions", "Sq", "AlgWrapper", "AlternatingAlgWrapper", "SqAlgWrapper", "FromArchive"]
 
 import abc
 import math
@@ -49,6 +49,7 @@ def _evalWrapper(eval_id: int, fn: Callable, *args, **kwargs) -> Tuple[int, floa
     exc = None
     fitness = None
     features = None
+    res = None
     try:
         res = fn(*args, **kwargs)
     except Exception as e:
@@ -65,6 +66,7 @@ def _severalEvalsWrapper(eval_id: int, fn: Callable, *args, **kwargs) -> Tuple[i
     exc = None
     fitness = None
     features = None
+    res = None
     try:
         res = fn(*args, **kwargs)
     except Exception as e:
@@ -134,6 +136,8 @@ class QDAlgorithmLike(Protocol):
     dimension: Optional[int]
     budget: int
 
+    def reinit(self) -> None: ...
+
     def ask(self) -> IndividualLike: ...
 
     def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
@@ -177,6 +181,9 @@ class QDAlgorithmLike(Protocol):
     def current_iter(self) -> int: ...
 
 
+
+
+
 # TODO say inspiration from nevergrad
 class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig):
     """TODO"""
@@ -227,12 +234,7 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
             else:
                 self._nb_objectives = len(self.container.fitness_domain)
 
-        self._nb_suggestions = 0
-        self._nb_suggestions_in_iteration = 0
-        self._nb_evaluations = 0
-        self._nb_evaluations_in_iteration = 0
-        self._nb_updated = 0
-        self._nb_updated_in_iteration = 0
+        self.reinit()
         self._callbacks = _CallbacksManager()
         self.optimisation_task = optimisation_task
         if add_default_logger:
@@ -259,11 +261,33 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
                 self._internal_tell(e, True)
 
 
+    def reinit(self) -> None:
+        self._nb_suggestions = 0
+        self._nb_suggestions_in_iteration = 0
+        self._nb_evaluations = 0
+        self._nb_evaluations_in_iteration = 0
+        self._nb_updated = 0
+        self._nb_updated_in_iteration = 0
+
+
     def __getstate__(self):
         odict = self.__dict__.copy()
         del odict['_base_ind_gen']
         return odict
 
+    # Do not copy the container when deep copying
+    def __deepcopy__(self, memo):
+        cls = self.__class__
+        result = cls.__new__(cls)
+        memo[id(self)] = result
+        for k, v in self.__dict__.items():
+            if k == "container":
+                setattr(result, k, v)
+            else:
+                setattr(result, k, copy.deepcopy(v, memo))
+        return result
+
+
     @property
     def optimisation_task(self) -> Union[str, Sequence[float]]:
         return self._optimisation_task
@@ -382,7 +406,8 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
 
 
     def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
-            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None) -> bool:
+            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None,
+            tell_container = True) -> bool:
         """Give the algorithm an evaluated `individual` (with a valid fitness function)
         and store it into the container.
 
@@ -407,19 +432,39 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
         if len(ind.fitness.values) != self._nb_objectives:
             raise ValueError(f"`individual` must possess a `fitness` attribute of dimension {self._nb_objectives} (specified through parameter `nb_objectives` when creating a QDAlgorithm sub-class), instead of {len(individual.fitness.values)}.")
         if features is not None:
-            ind.features = features
+            if isinstance(features, FeaturesLike):
+                ind.features = features
+            else:
+                ind.features.values = features
         if elapsed is not None:
             ind.elapsed = elapsed
-        added: bool = self.container.update([ind]) == 1
-        self._tell_after_container_update(ind, added)
+        xattr: Mapping[str, Any]
+        allowed_to_tell: bool
+        xattr, allowed_to_tell = self._tell_before_container_update(ind)
+        added: bool = self.container.update([ind]) == 1 if tell_container and allowed_to_tell else False
+        self._tell_after_container_update(ind, added, xattr)
         return added
 
 
-    def _tell_after_container_update(self, individual: IndividualLike, added: bool) -> None:
+    def _tell_before_container_update(self, individual: IndividualLike) -> Tuple[MutableMapping[str, Any], bool]:
+        """Called by the ``tell`` method before adding the individual to the container. Can be used to compile
+        a dictionary of information about the individual or container before the update operation occurs.
+        Return a dictionary of extended attributes and a boolean flag indicating whether the algorithm is allowed to tell the
+        individual or not.
+
+        Parameters
+        ----------
+        :param individual: IndividualLike
+            The `individual` provided to the algorithm.
+        """
+        return {}, True
+
+
+    def _tell_after_container_update(self, individual: IndividualLike, added: bool, xattr: Mapping[str, Any] = {}) -> None:
         if added:
             self._nb_updated += 1
             self._nb_updated_in_iteration += 1
-        self._internal_tell(individual, added)
+        self._internal_tell(individual, added, xattr)
         self._nb_evaluations += 1
         self._nb_evaluations_in_iteration += 1
         # Call callback functions
@@ -427,8 +472,8 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
             fn(self, individual)
 
 
-    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
-        """Internal tell method. To be overrode by subclasses, or left empty.
+    def _internal_tell(self, individual: IndividualLike, added_to_container: bool, xattr: Mapping[str, Any] = {}) -> None:
+        """Internal tell method. To be overridden by subclasses, or left empty.
         
         Parameters
         ----------
@@ -466,7 +511,7 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
 
 
     def _internal_best(self) -> IndividualLike:
-        """Return the individual with the best quality. To be overrode by subclasses.
+        """Return the individual with the best quality. To be overridden by subclasses.
         By default, returns the best individual of `self.container`."""
         return self.container.best # Assume `self.container` is not empty
 
@@ -492,10 +537,12 @@ class QDAlgorithm(abc.ABC, QDAlgorithmLike, Summarisable, Saveable, Copyable, Cr
             send_several_suggestions_to_fn: bool = False) -> IndividualLike:
         """TODO"""
         optimisation_start_time: float = timer()
+        # Init budget
         if budget is not None:
             self.budget = budget
         if self.budget is None:
             raise ValueError("`budget` must be provided.")
+
         _executor: ExecutorLike = executor if executor is not None else SequentialExecutor()
 
         # Call callback functions
@@ -650,7 +697,7 @@ class Lambda(QDAlgorithm):
         else:
             return self._ask_fn(base_ind) # type: ignore
 
-    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
+    def _internal_tell(self, individual: IndividualLike, added_to_container: bool, xattr: Mapping[str, Any] = {}) -> None:
         if self._tell_fn is not None:
             self._tell_fn(individual, added_to_container)
 
@@ -683,6 +730,154 @@ class AlgWithBatchSuggestions(QDAlgorithm):
 
 
 
+#class EmitterManager(QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig):
+class AlgWrapper(QDAlgorithm):
+    """Algorithm that contains and wraps other algorithms."""
+    algorithms: Sequence[QDAlgorithm]
+    current_idx: int = 0
+    current: QDAlgorithm
+    tell_container_when_switching: Union[bool, str]
+    batch_mode: bool
+
+    def __init__(self, algorithms: Any, container: Optional[Container] = None, budget: Optional[int] = None,
+            tell_container_when_switching: Union[bool, str] = False,
+            name: Optional[str] = None, batch_mode: bool = False, **kwargs):
+        if len(algorithms) == 0:
+            raise ValueError("`algorithms` must not be empty.")
+#        if isinstance(algorithms, Mapping):
+#            self.algorithms = []
+#            for k,v in algorithms.items():
+#                for i in range(v):
+#                    self.algorithms.append(k)
+#        else:
+#            self.algorithms = list(algorithms)
+        self.algorithms = list(algorithms)
+        self.current_idx = 0
+        self.current = self.algorithms[self.current_idx]
+        self.batch_mode = batch_mode
+
+        container = algorithms[0].container if container is None else container
+        budget_: int = sum((algo.budget for algo in self.algorithms)) if budget is None else budget
+        super().__init__(container, budget_, **kwargs)
+        delattr(self, 'container')
+
+        if isinstance(tell_container_when_switching, str):
+            if tell_container_when_switching not in ["True", "False", "only_best"]:
+                raise ValueError("`tell_container_when_switching` must be a bool or one of these strings: 'True', 'False', 'only_best'.")
+        self.tell_container_when_switching = tell_container_when_switching
+        self.name = name if name is not None else f"{self.__class__.__name__}-{id(self)}"
+        self.switch_to(0)
+
+
+    def __getattr__(self, attr):
+        if attr in self.__dict__:
+            return getattr(self, attr)
+        if "current" not in self.__dict__:
+            raise AttributeError
+        return getattr(self.current, attr)
+
+
+#    def __get_saved_state__(self) -> Mapping[str, Any]:
+#        """Return a dictionary containing the relevant information to save. Only include information from `self.algorithms`."""
+#        res = copy.deepcopy(self.current.__get_saved_state__())
+#        entries = {k:v for k,v in inspect.getmembers(self) if not k.startswith('_') and not inspect.ismethod(v)}
+#        return {**entries, **res}
+
+
+    @abc.abstractmethod
+    def next(self) -> None:
+        """Switch to the next algorithm in Sequence `self.algorithms`, if there is one."""
+        raise NotImplementedError("Method `_internal_ask` must be implemented by AlgWrapper subclasses.")
+#        if self.current_idx < len(self.algorithms) - 1:
+#            self.switch_to(self.current_idx + 1)
+#        else:
+#            self.switch_to(0)
+
+    def switch_to(self, idx: int):
+        """Switch to another algorithm of index `idx` in `self.algorithms`."""
+        assert idx < len(self.algorithms), f"`idx` must be < len(self.algorithms)."
+        #print(f"SWITCH to {idx}")
+        previous = self.current
+        self.current_idx = idx
+        self.current = self.algorithms[idx]
+        if (isinstance(self.tell_container_when_switching, bool) and self.tell_container_when_switching) or \
+                self.tell_container_when_switching == "True":
+            self.current.tell_container_entries()
+        elif self.tell_container_when_switching == "only_best":
+            try:
+                self.current.tell(previous.best()) # Exception when no individual is present
+            except Exception:
+                pass
+
+    def best(self) -> IndividualLike:
+        return self.current.best()
+
+#    def ask(self) -> IndividualLike:
+#        ind = self.current.ask()
+#        self.next()
+#        return ind
+
+    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
+        if self.batch_mode:
+            batch_size = self.current.batch_size
+            if np.isinf(batch_size):
+                batch_size = self.batch_size
+            if np.isinf(batch_size):
+                batch_size = 1
+            if self.current.nb_suggestions % self.current.batch_size == 0:
+                self.next()
+        else:
+            self.next()
+        ind = self.current.ask()
+        return ind
+
+    def tell(self, individual: IndividualLike, fitness: Optional[Any] = None,
+            features: Optional[FeaturesLike] = None, elapsed: Optional[float] = None,
+            tell_container = False) -> bool:
+        added: bool = self.current.tell(individual, fitness, features, elapsed)
+        self._tell_after_container_update(individual, added)
+        return added
+
+#    def optimise(self, evaluate: Callable, budget: Optional[int] = None, 
+#            batch_mode: bool = True, executor: Optional[ExecutorLike] = None,
+#            send_several_suggestions_to_fn: bool = False) -> IndividualLike:
+#        for _ in range(self.current_idx, len(self.algorithms)):
+#            try:
+#                res = self.current.optimise(evaluate, budget, batch_mode, executor, send_several_suggestions_to_fn)
+#            except Exception as e:
+#                warnings.warn(f"Optimisation failed with algorithm '{self.current.name}': {e}")
+#                traceback.print_exc()
+#            self.next()
+#        return res
+
+#    def add_callback(self, event: str, fn: Callable) -> None:
+#        for a in self.algorithms:
+#            a.add_callback(event, fn)
+
+
+
+
+@registry.register
+class AlternatingAlgWrapper(AlgWrapper):
+    def next(self) -> None:
+        """Switch to the next algorithm in Sequence `self.algorithms`, if there is one."""
+        if self.current_idx < len(self.algorithms) - 1:
+            self.switch_to(self.current_idx + 1)
+        else:
+            self.switch_to(0)
+
+
+@registry.register
+class SqAlgWrapper(AlgWrapper):
+    def next(self) -> None:
+        """Switch to the next algorithm in Sequence `self.algorithms`, if there is one."""
+        if self.current.nb_suggestions >= self.current.budget:
+            if self.current_idx < len(self.algorithms) - 1:
+                self.switch_to(self.current_idx + 1)
+
+
+
+
 @registry.register
 class Sq(QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig):
     """Sequence of algorithms. Useful to easily optimise several algorithms sequentially.  """
@@ -769,6 +964,9 @@ class Sq(QDAlgorithmLike, Summarisable, Saveable, Copyable, CreatableFromConfig)
             a.add_callback(event, fn)
 
 
+
+
+
 @registry.register
 class FromArchive(QDAlgorithm):
     """TODO"""
@@ -791,6 +989,9 @@ class FromArchive(QDAlgorithm):
         return base_ind
 
 
+
+
+
 # MODELINE	"{{{1
 # vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
 # vim:foldmethod=marker
