@@ -32,6 +32,11 @@ import copy
 from inspect import signature
 from functools import partial
 import warnings
+import numpy as np
+
+import os
+import psutil
+import signal
 
 from qdpy.utils import *
 
@@ -40,57 +45,6 @@ from qdpy.utils import *
 T = TypeVar("T")
 DomainLike = Tuple[float, float]
 ShapeLike = Tuple[int, ...]
-#BackendLike = Union[MutableSet[T], MutableSequence[T]]
-FitnessValuesLike = Sequence[T]
-FeaturesLike = Sequence
-
-@runtime
-class FitnessLike(Protocol):
-    """Fitness protocol inspired from (and compatible with) DEAP Fitness class."""
-    weights: FitnessValuesLike
-    def dominates(self, other: Any, obj: Any = slice(None)) -> bool: ...
-    def getValues(self) -> FitnessValuesLike: ...
-    def setValues(self, values: FitnessValuesLike) -> None: ...
-    def delValues(self) -> None: ...
-    @property
-    def values(self) -> FitnessValuesLike: ...
-    @values.setter
-    def values(self, values: FitnessValuesLike) -> None: ...
-    @values.deleter
-    def values(self) -> None: ...
-    @property
-    def valid(self) -> bool: ...
-#    @overload
-#    def __getitem__(self, i: int) -> Any: ...
-#    @overload
-#    def __getitem__(self, s: slice) -> Sequence[Any]: ...
-#    def __len__(self) -> int: ...
-
-
-@runtime
-class IndividualLike(Protocol):
-    name: str
-    fitness: FitnessLike
-    features: FeaturesLike
-    elapsed: float
-    def dominates(self, other: Any) -> bool: ...
-    def __setitem__(self, key, values) -> None: ...
-
-
-#@runtime
-#class DEAPFitnessLike(FitnessLike):
-#    pass
-#
-#@runtime
-#class DEAPIndividualLike(Protocol):
-#    fitness: DEAPStyleFitnessLike
-#    features: FeaturesLike
-
-
-#FitnessLike = Sequence
-FitnessGetter = Callable[[T], FitnessLike]
-FeaturesGetter = Callable[[T], FeaturesLike]
-GridIndexLike = ShapeLike
 
 
 ########### CONCURRENCY ########### {{{1
@@ -207,12 +161,26 @@ try:
     def _ray_deploy_func(func, kwargs, *args):
         return func(*args, **kwargs)
 
+    @ray.remote
+    class RayDeployment(object):
+        def __init__(self, i):
+            pass # TODO
+        def ray_deploy_func(self, func, kwargs, *args):
+            return func(*args, **kwargs)
+
     class RayExecutor(ExecutorLike): # type: ignore
         """Executor that encapsulate ray concurrency functions. Need the ray package to be installed and importable."""
         def __init__(self) -> None:
-            ray.init(ignore_reinit_error=True)
+            num_cpus = psutil.cpu_count(logical=True)
+            ray.init(num_cpus=num_cpus, ignore_reinit_error=True)
+            #self.nb_actors = num_cpus
+            #self.actors = [RayDeployment.remote(i) for i in range(self.nb_actors)]
+            #self.current_actor_idx = 0
         def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> FutureLike:
             return RayFuture(_ray_deploy_func.remote(fn, kwargs, *args)) # TODO
+            #future = RayFuture(self.actors[self.current_actor_idx].ray_deploy_func.remote(fn, kwargs, *args))
+            #self.current_actor_idx = (self.current_actor_idx + 1) % self.nb_actors
+            #return future
         def map(self, fn: Callable, *iterables: Any) -> Iterable:
             # From cpython code 'concurrent/futures/_base.py'
             fs = [self.submit(fn, *args) for args in zip(*iterables)]
@@ -230,7 +198,7 @@ try:
                         future.cancel()
             return result_iterator()
         def shutdown(self, wait: bool = True) -> None:
-            pass # TODO
+            ray.shutdown()
 
     def _ray_as_completed_fn(fs, timeout=None):
         oid_lst = [x.oid for x in fs]
@@ -286,6 +254,26 @@ def generic_as_completed(fs: Sequence[FutureLike], timeout: Optional[int]=None)
 
 
 
+
+# Hack to terminate child process/threads gracefully using Ctrl-C
+# From https://stackoverflow.com/questions/32160054/keyboard-interrupts-with-pythons-multiprocessing-pool-and-map-function/45259908
+_parent_id = os.getpid()
+def _worker_init():
+#    signal.signal(signal.SIGINT, signal.SIG_IGN)
+    def sig_int(signal_num, frame):
+        #print('signal: %s' % signal_num)
+        parent = psutil.Process(_parent_id)
+        for child in parent.children():
+            if child.pid != os.getpid():
+                print("killing child: %s" % child.pid)
+                child.kill()
+        #print("killing parent: %s" % _parent_id)
+        parent.kill()
+        #print("suicide: %s" % os.getpid())
+        psutil.Process(os.getpid()).kill()
+    signal.signal(signal.SIGINT, sig_int)
+
+
 # TODO Support for Dask
 class ParallelismManager(object):
     """Simplify parallelism handling by generating on-the-fly executor and toolboxes to conform to a
@@ -299,13 +287,15 @@ class ParallelismManager(object):
     max_workers: Optional[int]
     _parallelism_type: str = ""
     _parallelism_type_list: Sequence[str] = ['none', 'sequential', 'multiprocessing', 'concurrent', 'multithreading', 'scoop', 'ray']
+    force_worker_suicide: bool
 
     def __init__(self, parallelism_type: str = "multiprocessing", max_workers: Optional[int] = None,
-            toolbox: Optional[Any] = None) -> None:
+            toolbox: Optional[Any] = None, force_worker_suicide: bool = True) -> None:
         self.parallelism_type = parallelism_type
         self.max_workers = max_workers
         self.orig_toolbox = toolbox
         self.toolbox = copy.deepcopy(toolbox) if toolbox is not None else None
+        self.force_worker_suicide = force_worker_suicide
 
     @property
     def parallelism_type(self) -> str:
@@ -329,6 +319,10 @@ class ParallelismManager(object):
 
     def open(self) -> Any:
         """Open and prepare a new executor according to `self.parallelism_type`."""
+        if self.force_worker_suicide:
+            init_fn: Optional[Callable] = _worker_init
+        else:
+            init_fn = None
         if self.parallelism_type == "none" or self.parallelism_type == "sequential":
             self.executor = SequentialExecutor()
         elif self.parallelism_type == "multiprocessing" or self.parallelism_type == "concurrent":
@@ -336,7 +330,7 @@ class ParallelismManager(object):
             if sys.version_info >= (3, 7):
                 import multiprocessing
                 ctx = multiprocessing.get_context("forkserver")
-                self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=ctx) # type: ignore
+                self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers, mp_context=ctx, initializer=init_fn) # type: ignore
             else:
                 self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.max_workers) # type: ignore
         elif self.parallelism_type == "multithreading":
@@ -360,175 +354,6 @@ class ParallelismManager(object):
 
 
 
-
-########### BASE OPTIMISATION CLASSES ########### {{{1
-
-class Fitness(FitnessLike, Sequence[Any]):
-    """Fitness implementation inspired from DEAP Fitness class. It can be used without problem with most (propably all) DEAP methods."""
-
-    weights: FitnessValuesLike = ()
-    wvalues: FitnessValuesLike = ()
-
-    def __new__(cls, values: FitnessValuesLike=(), weights: Optional[FitnessValuesLike]=None):
-        return super(Fitness, cls).__new__(cls)
-
-    def __init__(self, values: FitnessValuesLike=(), weights: Optional[FitnessValuesLike]=None) -> None:
-        if weights is None:
-            self.weights = tuple([-1.0 for _ in range(len(values))]) # Defaults to minimisation
-        else:
-            self.weights = weights
-        if len(self.weights) != len(values):
-            raise ValueError("``values`` and ``weights`` must have the same length.")
-        self.values = values
-
-    @property
-    def values(self) -> FitnessValuesLike:
-        return tuple(map(truediv, self.wvalues, self.weights))
-
-    @values.setter
-    def values(self, values: FitnessValuesLike) -> None:
-        try:
-            self.wvalues = tuple(map(mul, values, self.weights))
-        except TypeError:
-            raise ValueError("Invalid ``value`` parameter.")
-
-    @values.deleter
-    def values(self) -> None:
-        self.wvalues = ()
-
-    def getValues(self) -> FitnessValuesLike:
-        return self.values
-
-    def setValues(self, values: FitnessValuesLike) -> None:
-        self.values = values
-
-    def delValues(self) -> None:
-        del self.values
-
-    # FROM DEAP
-    def dominates(self, other: Any, obj: Any = slice(None)) -> bool:
-        """Return true if each objective of ``self`` is not strictly worse than
-        the corresponding objective of ``other`` and at least one objective is
-        strictly better.
-        """
-#    :param obj: Slice indicating on which objectives the domination is
-#                tested. The default value is `slice(None)`, representing
-#                every objectives.  """
-        not_equal: bool = False
-        for self_wvalue, other_wvalue in zip(self.wvalues[obj], other.wvalues[obj]):
-            if self_wvalue > other_wvalue:
-                not_equal = True
-            elif self_wvalue < other_wvalue:
-                return False
-        return not_equal
-
-    @property
-    def valid(self) -> bool:
-        """Assess if a fitness is valid or not."""
-        return len(self.wvalues) != 0
-
-    def __len__(self) -> int:
-        return len(self.wvalues)
-
-#    @overload
-#    def __getitem__(self, index: int) -> T: ...
-#
-#    @overload
-#    def __getitem__(self, s: slice) -> Sequence: ...
-
-    def __getitem__(self, key):
-        return self.values[key]
-
-    def __contains__(self, key: Any) -> bool:
-        return key in self.values
-
-    def __iter__(self) -> Iterator:
-        return iter(self.values)
-
-    def __reversed__(self) -> Iterator:
-        return reversed(self.values)
-
-
-    def __hash__(self) -> int:
-        return hash(self.wvalues)
-
-    def __gt__(self, other: Any) -> bool:
-        return not self.__le__(other)
-
-    def __ge__(self, other: Any) -> bool:
-        return not self.__lt__(other)
-
-    def __le__(self, other: Any) -> bool:
-        return self.wvalues <= other.wvalues
-
-    def __lt__(self, other: Any) -> bool:
-        return self.wvalues < other.wvalues
-
-    def __eq__(self, other: Any) -> bool:
-        return self.wvalues == other.wvalues
-
-    def __ne__(self, other: Any) -> bool:
-        return not self.__eq__(other)
-
-    def __str__(self) -> str:
-        return str(self.values if self.valid else tuple())
-
-    def __repr__(self) -> str:
-        if not self:
-            return "%s()" % (self.__class__.__name__,)
-        return "%s(%r)" % (self.__class__.__name__, list(self))
-
-
-class Individual(list, IndividualLike):
-    """Qdpy Individual class. Note that containers and algorithms all use internally either the QDPYIndividualLike Protocol or the IndividualWrapper class, so you can easily declare an alternative class to Individual. TODO""" # TODO
-
-    name: str
-    fitness: FitnessLike
-    features: FeaturesLike
-    elapsed: float = math.nan
-
-    def __init__(self, iterable: Optional[Iterable] = None,
-            name: Optional[str] = None,
-            fitness: Optional[FitnessLike] = None, features: FeaturesLike = ()) -> None:
-        if iterable is not None:
-            self.extend(iterable)
-        self.name = name if name else ""
-        self.fitness = fitness if fitness is not None else Fitness()
-        self.features = features
-
-    def __repr__(self) -> str:
-        if not self:
-            return "%s()" % (self.__class__.__name__,)
-        return "%s(%r)" % (self.__class__.__name__, list(self))
-
-    def dominates(self, other: Any) -> bool:
-        """Return true if ``self`` dominates ``other``. """
-        return self.fitness.dominates(other.fitness)
-
-    # TODO : improve performance ! (quick and dirty solution !)
-    def __hash__(self):
-        return hash(tuple(self))
-
-    def __eq__(self, other):
-        return (self.__class__ == other.__class__ and tuple(self) == tuple(other))
-
-
-
-
-## TODO
-#class IndividualWrapper(Individual):
-#    """TODO""" # TODO
-#
-#    AnyIndividualLike = Union[QDPYIndividualLike, DEAPStyleIndividualLike]
-#
-#    _ind: AnyIndividualLike
-#    fitness: FitnessLike
-#    features: FeaturesLike
-#
-#    def __init__(self, individual: AnyIndividualLike)
-
-
-
 ########### BASE CLASSES ########### {{{1
 class Summarisable(object):
     """Describes a class that can be summarised by using the `self.summary` method.
@@ -537,7 +362,7 @@ class Summarisable(object):
 
     def __get_summary_state__(self) -> Mapping[str, Any]:
         """Return a dictionary containing the relevant entries to build a summary of the class.
-        By default, it includes all public attributes of the class. Must be overrode by subclasses."""
+        By default, it includes all public attributes of the class. Must be overridden by subclasses."""
         # Find public attributes
         entries = {}
         for k, v in inspect.getmembers(self):
@@ -596,7 +421,7 @@ class Saveable(object):
 
     def __get_saved_state__(self) -> Mapping[str, Any]:
         """Return a dictionary containing the relevant information to save.
-        By default, it includes all public attributes of the class. Must be overrode by subclasses."""
+        By default, it includes all public attributes of the class. Must be overridden by subclasses."""
         # Find public attributes
         entries = {k:v for k,v in inspect.getmembers(self) if not k.startswith('_') and not inspect.ismethod(v)}
         return entries
@@ -723,10 +548,21 @@ class Factory(dict):
                 config[k] = sub_obj
             elif isinstance(v, str) and v in self:
                 config[k] = self[v]
+            elif isinstance(v, Mapping):
+                v_ = copy.copy(v)
+                config[k] = []
+                for k2, v2 in v_.items():
+                    if isinstance(k2, str) and isinstance(v2, int) and k2 in self:
+                        for _ in range(v2):
+                            new_k2 = copy.deepcopy(self[k2])
+                            new_k2.reinit()
+                            config[k].append(new_k2)
             elif is_iterable(v):
                 for i, val in enumerate(v):
                     if isinstance(val, str) and val in self:
-                        config[k][i] = self[val]
+                        new_val = copy.deepcopy(self[val])
+                        new_val.reinit()
+                        config[k][i] = new_val
             if not is_iterable(v) or (is_iterable(v) and "type" not in v):
                 default_config[k] = config[k]
 
@@ -778,6 +614,8 @@ class Factory(dict):
 
 
 
+from qdpy.phenotype import *
+
 # MODELINE	"{{{1
 # vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
 # vim:foldmethod=marker
