@@ -30,6 +30,7 @@ import numpy as np
 from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, MutableMapping, Mapping, overload
 from typing_extensions import runtime, Protocol
 import traceback
+import random
 
 from qdpy.utils import *
 from qdpy.base import *
@@ -268,7 +269,7 @@ class Container(Sequence, Summarisable, Copyable, CreatableFromConfig):
         maxima = np.array(self[0].features) # type: ignore
         minima = np.array(self[0].features) # type: ignore
         for ind in self:
-            ind_f = ind.features
+            ind_f = ind.features.values
             for i in range(len(maxima)):
                 if ind_f[i] > maxima[i]:
                     maxima[i] = ind_f[i]
@@ -629,14 +630,17 @@ class Grid(Container):
     recentness_per_bin: GridRecentnessPerBinLike
     history_recentness_per_bin: GridRecentnessPerBinLike
     activity_per_bin: np.array
+    discard_random_on_bin_overload: bool
 
 
     def __init__(self, iterable: Optional[Iterable] = None,
             shape: Union[ShapeLike, int] = (1,), max_items_per_bin: int = 1,
             fitness_domain: Optional[Sequence[DomainLike]] = ((0., np.inf),),
+            discard_random_on_bin_overload = False,
             **kwargs: Any) -> None:
         self._shape = tuplify(shape)
         self._max_items_per_bin = max_items_per_bin
+        self.discard_random_on_bin_overload = discard_random_on_bin_overload
         super().__init__([], fitness_domain=fitness_domain, **kwargs)
         if self.features_domain is None or len(self.features_domain) == 0:
             raise ValueError("`features_domain` must be specified and have a length > 0.")
@@ -729,14 +733,14 @@ class Grid(Container):
     def index_grid(self, features: FeaturesLike) -> GridIndexLike:
         """Get the index in the grid of a given individual with features ``features``, raising an IndexError if it is outside the grid. """
         index: List[int] = []
-        if len(features) != len(self.shape):
-            raise IndexError(f"Length of parameter ``features`` ({len(features)}) does not corresponds to the number of dimensions of the grid ({len(self.shape)}).")
-        for i in range(len(features)):
-            normalised_feature: float = features[i] - self.features_domain[i][0]
+        if len(features.values) != len(self.shape):
+            raise IndexError(f"Length of parameter ``features.values`` ({len(features.values)}) does not corresponds to the number of dimensions of the grid ({len(self.shape)}).")
+        for i in range(len(features.values)):
+            normalised_feature: float = features.values[i] - self.features_domain[i][0]
             if normalised_feature == self.features_domain[i][1] - self.features_domain[i][0]:
                 partial: int = self.shape[i] - 1
             elif normalised_feature > self.features_domain[i][1] - self.features_domain[i][0]:
-                raise IndexError(f"``features`` ({str(features)}) out of bounds ({str(self.features_domain)})")
+                raise IndexError(f"``features.values`` ({str(features.values)}) out of bounds ({str(self.features_domain)})")
             else:
                 partial = int(normalised_feature / self._bins_size[i])
             index.append(partial)
@@ -789,18 +793,24 @@ class Grid(Container):
         if self._nb_items_per_bin[ig] < self.max_items_per_bin:
             can_be_added = True
         else:
-            worst_idx = 0
-            worst: IndividualLike = self.solutions[ig][worst_idx]
-            if self._nb_items_per_bin[ig] > 1:
-                for i in range(1, self._nb_items_per_bin[ig]):
-                    s = self.solutions[ig][i]
-                    if worst.fitness.dominates(s.fitness):
-                        worst = s
-                        worst_idx = i
-            if individual.fitness.dominates(worst.fitness):
-                Container.discard(self, self.solutions[ig][worst_idx])
-                self._discard_from_grid(ig, worst_idx)
+            if self.discard_random_on_bin_overload:
+                idx_to_discard = random.randint(0, len(self.solutions[ig])-1)
+                Container.discard(self, self.solutions[ig][idx_to_discard])
+                self._discard_from_grid(ig, idx_to_discard)
                 can_be_added = True
+            else:
+                worst_idx = 0
+                worst: IndividualLike = self.solutions[ig][worst_idx]
+                if self._nb_items_per_bin[ig] > 1:
+                    for i in range(1, self._nb_items_per_bin[ig]):
+                        s = self.solutions[ig][i]
+                        if worst.fitness.dominates(s.fitness):
+                            worst = s
+                            worst_idx = i
+                if individual.fitness.dominates(worst.fitness):
+                    Container.discard(self, self.solutions[ig][worst_idx])
+                    self._discard_from_grid(ig, worst_idx)
+                    can_be_added = True
 
         # Add individual in grid, if there are enough empty spots
         if can_be_added:
